@@ -75,63 +75,90 @@ __BEGIN_TEMPLATE_ARGS__
 1
 ,P1
 ,P1
-*(args++)=(unsigned long)(P1);
+*(__args++)=(unsigned long)(P1);
 __END_TEMPLATE_ARGS__
 __BEGIN_TEMPLATE_ARGS__
 2
 ,P1,P2
 ,P1,P2
-*(args++)=(unsigned long)(P1);*(args++)=(unsigned long)(P2);
+*(__args++)=(unsigned long)(P1);*(__args++)=(unsigned long)(P2);
 __END_TEMPLATE_ARGS__
 __BEGIN_TEMPLATE_ARGS__
 3
 ,P1,P2,P3
 ,P1,P2,P3
-*(args++)=(unsigned long)(P1);*(args++)=(unsigned long)(P2);*(args++)=(unsigned long)(P3);
+*(__args++)=(unsigned long)(P1);*(__args++)=(unsigned long)(P2);*(__args++)=(unsigned long)(P3);
 __END_TEMPLATE_ARGS__
 __BEGIN_TEMPLATE_ARGS__
 4
 ,P1,P2,P3,P4
 ,P1,P2,P3,P4
-*(args++)=(unsigned long)(P1);*(args++)=(unsigned long)(P2);*(args++)=(unsigned long)(P3);*(args++)=(unsigned long)(P4);
+*(__args++)=(unsigned long)(P1);*(__args++)=(unsigned long)(P2);*(__args++)=(unsigned long)(P3);*(__args++)=(unsigned long)(P4);
 __END_TEMPLATE_ARGS__
 __BEGIN_TEMPLATE_ARGS__
 5
 ,P1,P2,P3,P4,P5
 ,P1,P2,P3,P4,P5
-*(args++)=(unsigned long)(P1);*(args++)=(unsigned long)(P2);*(args++)=(unsigned long)(P3);*(args++)=(unsigned long)(P4);*(args++)=(unsigned long)(P5);
+*(__args++)=(unsigned long)(P1);*(__args++)=(unsigned long)(P2);*(__args++)=(unsigned long)(P3);*(__args++)=(unsigned long)(P4);*(__args++)=(unsigned long)(P5);
 __END_TEMPLATE_ARGS__
 __BEGIN_TEMPLATE_ARGS__
 6
 ,P1,P2,P3,P4,P5,P6
 ,P1,P2,P3,P4,P5,P6
-*(args++)=(unsigned long)(P1);*(args++)=(unsigned long)(P2);*(args++)=(unsigned long)(P3);*(args++)=(unsigned long)(P4);*(args++)=(unsigned long)(P5);*(args++)=(unsigned long)(P6);
+*(__args++)=(unsigned long)(P1);*(__args++)=(unsigned long)(P2);*(__args++)=(unsigned long)(P3);*(__args++)=(unsigned long)(P4);*(__args++)=(unsigned long)(P5);*(__args++)=(unsigned long)(P6);
+__END_TEMPLATE_ARGS__
+__BEGIN_TEMPLATE_ARGS__
+7
+,P1,P2,P3,P4,P5,P6,P7
+,P1,P2,P3,P4,P5,P6,P7
+*(__args++)=(unsigned long)(P1);*(__args++)=(unsigned long)(P2);*(__args++)=(unsigned long)(P3);*(__args++)=(unsigned long)(P4);*(__args++)=(unsigned long)(P5);*(__args++)=(unsigned long)(P6);*(__args++)=(unsigned long)(P7);
 __END_TEMPLATE_ARGS__
 
 #define FUT_CODE(CODE, n) ((((unsigned long)(CODE))<<8) | ((n) + 1))
 #define FUT_SIZE(n) (sizeof(fxt_trace_user_raw_t)+(n)*sizeof(long))
 #define FUT_STRLEN(s) ((strlen(s)+sizeof(long)-1)/sizeof(long))
 
+#define FUT_NEEDS_COMMIT
+
 __BEGIN_TEMPLATE__
 #define FUT_RAW_PROBE__NUMBER__(CODE__DECLARE_ARGS__) do {		\
 		if(fut_active) {					\
-			unsigned long *args __attribute__((unused))=	\
+			unsigned long *__args __attribute__((unused))=	\
 				fut_getstampedbuffer(CODE,		\
 						     FUT_SIZE(__NUMBER__)); \
 			__ARGS_REGISTER__				\
-				}					\
+			fut_commitstampedbuffer(FUT_SIZE(__NUMBER__)); \
+		}							\
+	} while (0)
+
+#define FUT_RAW_PROBE__NUMBER__STR(CODE__DECLARE_ARGS__,S) do {		\
+		if(fut_active) {					\
+			const char *__s = (S);				\
+			size_t __slen = strlen(__s);			\
+			size_t __max = (FXT_MAX_PARAMS - __NUMBER__) * sizeof(unsigned long); \
+			size_t __len = __slen > __max ? __max : __slen;	\
+			unsigned __nbargs_str = (__len + sizeof(unsigned long) - 1) / sizeof (unsigned long); \
+			unsigned __nbargs = __NUMBER__ + __nbargs_str;	\
+			unsigned long *__args =				\
+				fut_getstampedbuffer(FUT_CODE(CODE, __nbargs), FUT_SIZE(__nbargs)); \
+			__ARGS_REGISTER__				\
+			memcpy(__args, __s, __len);				\
+			memset((void*)__args+__len, 0, __nbargs_str*sizeof(long)-__len);	\
+			fut_commitstampedbuffer(FUT_SIZE(__nbargs)); \
+		}							\
 	} while (0)
 
 __END_TEMPLATE__
 #define FUT_RAW_PROBESTR(CODE,S) do {					\
 		if(fut_active) {					\
-			const char *s = (S);				\
-			size_t len = strlen(s);				\
-			size_t n = FUT_STRLEN(s);			\
-			void *args __attribute__((unused)) =		\
-				fut_getstampedbuffer(CODE, FUT_SIZE(n)); \
-			memcpy(args, s, len);				\
-			memset(args+len, 0, n*sizeof(long)-len);	\
+			const char *__s = (S);				\
+			size_t __len = strlen(__s);			\
+			size_t __n = FUT_STRLEN(__s);			\
+			unsigned long *__args __attribute__((unused)) =	\
+				fut_getstampedbuffer(CODE, FUT_SIZE(__n)); \
+			memcpy(__args, __s, __len);			\
+			memset((void*)__args+__len, 0, __n*sizeof(long)-__len);	\
+			fut_commitstampedbuffer(FUT_SIZE(__n)); \
 		}							\
 	} while (0)
 
@@ -140,16 +167,22 @@ __BEGIN_TEMPLATE__
         FUT_RAW_PROBE__NUMBER__(FUT_CODE(CODE, __NUMBER__)__CALL_ARGS__); \
 } while (0)
 
+#define FUT_DO_PROBE__NUMBER__STR(CODE__DECLARE_ARGS__,S) do { \
+        FUT_RAW_PROBE__NUMBER__STR(CODE __CALL_ARGS__,S); \
+} while (0)
+
 __END_TEMPLATE__
 #define FUT_DO_PROBESTR(CODE,S) do { \
-	FUT_RAW_PROBESTR(FUT_CODE(CODE, FUT_STRLEN(s)), S); \
+	FUT_RAW_PROBESTR(FUT_CODE(CODE, FUT_STRLEN(__s)), S); \
 } while (0)
 
 /******************************************************/
-/* NEVER call fut_getstampedbuffer directly           */
+/* NEVER call fut_getstampedbuffer or                 */
+/* fut_commitstampedbuffer directly                   */
 /* Interface can change without forewarning           */
 /******************************************************/
 extern unsigned long* fut_getstampedbuffer(unsigned long code, int size);
+extern unsigned long* fut_commitstampedbuffer(int size);
 
 /* This can be overriden by the user */
 extern uint64_t fut_getstamp(void);
@@ -160,7 +193,7 @@ __attribute__((no_instrument_function))
 #endif
 FUT_RAW_PROBE(unsigned long code, ...) {
 	if(fut_active) {		
-	int nb=(code&0xff)-1;
+	int nb=(code&0xff)-1, orig_nb = nb;
         unsigned long *args=fut_getstampedbuffer(code, FUT_SIZE(nb));
 	va_list list;
 
@@ -169,6 +202,7 @@ FUT_RAW_PROBE(unsigned long code, ...) {
 		*(args++)=va_arg(list, unsigned long);
 	}
 	va_end(list);
+	fut_commitstampedbuffer(FUT_SIZE(orig_nb));
 	} 
 }
 #define FUT_DO_PROBE(CODE, ...) do { \
@@ -305,7 +339,10 @@ __END_TEMPLATE__
 
 #ifndef __ASSEMBLY__
 extern unsigned long * volatile fut_next_slot;
+extern unsigned long * volatile fut_first_slot;
 extern unsigned long * volatile fut_last_slot;
+extern size_t volatile fut_filled_slot;
+
 extern pthread_spinlock_t fut_slot_lock;
 
 extern int fut_setup( unsigned int nints, unsigned int keymask,
