@@ -10,9 +10,11 @@
 #  endif
 #endif
 
+#undef NDEBUG
 #include <pthread.h>
 #include <unistd.h>
 #include <stdlib.h>
+#include <assert.h>
 
 #ifdef HAVE_CLOCK_GETTIME
 #include <time.h>
@@ -196,9 +198,33 @@ fut_header (unsigned long code, ... )
 /* poor man's cmpxchg,  supprimer lorsque tbx implmentera cmpxchg lui-mme */
 #define ma_cmpxchg(ptr,o,n) \
 	__sync_val_compare_and_swap((ptr), (o), (n))
+#define add_return(ptr,s) \
+	__sync_fetch_and_add((ptr), (s))
 #endif
 #endif
 
+#define i386_sync() __asm__ __volatile__("lock; addl $0,0(%%esp)" ::: "memory")
+#define x86_64_sync() __asm__ __volatile__("lock; addl $0,0(%%rsp)" ::: "memory")
+#define ppc_sync() __asm__ __volatile__("sync" ::: "memory")
+
+#if defined(__i386__) || defined(__KNC__) || defined(__KNF__)
+#define mb() i386_sync()
+#define rmb() i386_sync()
+#define wmb() i386_sync()
+#elif defined(__x86_64__)
+#define mb() x86_64_sync()
+#define rmb() __asm__ __volatile__("lfence" ::: "memory")
+#define wmb() __asm__ __volatile__("sfence" ::: "memory")
+#elif defined(__ppc__) || defined(__ppc64__)
+#define mb() ppc_sync()
+#define rmb() ppc_sync()
+#define wmb() ppc_sync()
+#else
+#define mb() __sync_synchronize()
+#define rmb() __sync_synchronize()
+#define wmb() __sync_synchronize()
+#endif
+
 static unsigned long trash_buffer[8];
 extern int allow_fut_flush;
 extern pthread_mutex_t fut_flush_lock;
@@ -256,11 +282,17 @@ unsigned long* fut_getstampedbuffer(unsigned long code,
 	unsigned long *prev_slot, *next_slot;
 	uint64_t stamp;
  retry:
-#if 1
+#if ! USE_SPINLOCKS
 	/* compare and swap method */
 	do {
-		prev_slot=(unsigned long *)fut_next_slot;
+		/* Wait for flusher to finish flushing */
+		do {
+			rmb();
+			prev_slot=(unsigned long*)fut_next_slot;
+		} while (prev_slot > fut_last_slot);
 		next_slot=(unsigned long*)((unsigned long)prev_slot+size);
+		/* Note: we don't want to use add_return here, as we want to
+		 * make sure we are not reserving beyond fut_last_slot */
 	} while (prev_slot != ma_cmpxchg(&fut_next_slot, 
 					 prev_slot, next_slot));
 	stamp = fut_getstamp();
@@ -276,7 +308,7 @@ unsigned long* fut_getstampedbuffer(unsigned long code,
 
 	if (tbx_unlikely(next_slot > fut_last_slot)) {
 	  if(allow_fut_flush) {
-	    /* flush the buffer to disk */
+	    /* our record doesn't fit, flush the buffer to disk */
 
 	    if (prev_slot > fut_last_slot)
 	      /* another thread is going to flush the buffer */
@@ -284,12 +316,16 @@ unsigned long* fut_getstampedbuffer(unsigned long code,
 
 	    pthread_mutex_lock(&fut_flush_lock);
 
-	    if ((prev_slot + size > fut_last_slot)
-		&& (prev_slot <= fut_last_slot)) {
-
-	      fut_flush(NULL, prev_slot, 1);
-	      __fut_reset_pointers();
-	    }
+	    /* Wait for all threads to commit their writes */
+	    while (fut_filled_slot < (void*) prev_slot - (void*) fut_first_slot)
+	      rmb();
+	    /* Make sure others have finished writing */
+	    rmb();
+	    assert(fut_filled_slot == (void*) prev_slot - (void*) fut_first_slot);
+	    fut_flush(NULL, prev_slot, 1);
+	    /* synchronize between the flush and other threads */
+	    mb();
+	    __fut_reset_pointers();
 
 	    pthread_mutex_unlock(&fut_flush_lock);
 
@@ -308,4 +344,17 @@ unsigned long* fut_getstampedbuffer(unsigned long code,
 	return __fut_record_event(rec, stamp, code);
 }
 
+void fut_commitstampedbuffer(int size)
+{
+#if !USE_SPINLOCKS
+	/* Make sure flusher sees our writes */
+	wmb();
+	add_return(&fut_filled_slot, size);
+#else
+	pthread_spin_lock(&fut_slot_lock);
+	fut_filled_slot += size;
+	pthread_spin_unlock(&fut_slot_lock);
+#endif
+}
+
 #endif /* PROFILE_NEW_FORMAT */
